import 'package:nexbounce/nexbounce.dart';

typedef Dispatcher<T> = Future<void> Function(Action<T> action);
typedef Subscriber<T> = void Function(T state);

class Subscription<T> {
  final Store<T> store;
  final Subscriber<T> subscriber;

  const Subscription({required this.store, required this.subscriber});

  void cancel() => store._subscribers.remove(subscriber);
}

/// A base class that has to be extended to create Actions.
abstract class Action<T> {
  const Action();

  Future<T> activate(T state);
}

abstract class CompoundAction<T> {
  const CompoundAction();

  Future<void> activate(T state, Dispatcher<T> dispatch);
}

/// A base class that has to be extended to create stores.
abstract class Store<T> {
  final _subscribers = <Subscriber<T>>{};
  final _debouncer = Nexbounce();
  final bool logger;

  T _state;

  Store(this._state, {this.logger = false});

  /// Returns the current state of store.
  T get state => _state;

  /// Dispatches an action to update the state.
  Future<void> dispatch(Action<T> action) async {
    final newState = await action.activate(state);

    if (logger) {
      print('''
store: $this
action: $action
old state: $state
new state: $newState\n''');
    }

    _state = newState;

    if (newState != state) _publish();
  }

  Future<void> dispatchCompound(CompoundAction<T> compoundAction) async {
    if (logger) print('init compound action: $compoundAction\n');

    await compoundAction.activate(state, dispatch);

    if (logger) print('end of compound action: $compoundAction\n');
  }

  /// Accepts a callback as its argument which is run after state change.
  /// Note that running callbacks is debounced.
  Subscription<T> subscribe(Subscriber<T> subscriber) {
    _subscribers.add(subscriber);
    return Subscription(store: this, subscriber: subscriber);
  }

  /// Same as `subscribe` but callback is immediately called for the first time.
  Subscription<T> runAndSubscribe(Subscriber<T> subscriber) {
    subscriber(state);
    return subscribe(subscriber);
  }

  void _publish() {
    _debouncer.enqueue(() {
      for (final subscriber in _subscribers) {
        subscriber(state);
      }
    });
  }
}
